from collections import namedtuple
import numpy as np
import matplotlib.pyplot as plt
from seismod1d.rockphysics import conversions


def backus_average_cij(rho_in, c11_in, c13_in, c33_in, c44_in, c66_in):

    """
    Backus (1962)
    Input must have same uniform depth sampling|

    """

    rho = np.mean(rho_in)
    c11 = np.mean(c11_in - (c13_in ** 2 / c33_in)) + (
        np.mean(c13_in / c33_in) ** 2 / np.mean(1 / c33_in)
    )
    c13 = np.mean(c13_in / c33_in) * (1 / np.mean(1 / c33_in))
    c33 = 1 / np.mean(1 / c33_in)
    c44 = 1 / np.mean(1 / c44_in)
    c66 = np.mean(c66_in)

    BackusAverage_cij = namedtuple(
        "BackusAverage_cij", ["rho", "c11", "c13", "c33", "c44", "c66"]
    )
    return BackusAverage_cij(rho, c11, c13, c33, c44, c66)


def constant_length(
    zz, vp, vs, rho, delta, epsilon, gamma, block_length=2, qc_plot=False
):

    # get number of samples per blocking length
    nz = len(zz)  # total number of samples
    dz = zz[1] - zz[0]  # assuming constant depth interval
    iz = int(np.round(block_length / dz))  # number of samples per block
    
    # add small number to vs to avoid divide by zero
    small_value = 1e-12
    vs[vs == 0] = small_value

    # convert elastic parameters to elastic moduli
    rho, c11, c13, c33, c44, c66 = conversions.conv_vti_thoms2cij(
        rho, vp, vs, delT=delta, epsT=epsilon, gamT=gamma
    )
    
    # Note: values at first depth sample are not used due to "down-to" assumption, 
    # meaning that these values are valid down to first depth sample

    # get indices of start and end of each blocks
    istart = np.arange(0, nz, iz)  + 1 # index of first sample of each block
    iend = istart[1:] - 1  # index of last sample of each block
    iend = np.append(iend, [nz - 1])  # add

    # ensure last block is not too small
    ilast = iend[-1] - istart[-1]
    if ilast < iz - 1:
        istart = np.delete(istart, -1)
        iend = np.delete(iend, -2)

    # - initiate blocks
    rho_b = np.empty(np.shape(istart))
    c11_b = np.empty(np.shape(istart))
    c13_b = np.empty(np.shape(istart))
    c33_b = np.empty(np.shape(istart))
    c44_b = np.empty(np.shape(istart))
    c66_b = np.empty(np.shape(istart))

    # loop through all data and create blocks
    for ii in range(len(istart)):

        # select indexes for selected block
        jj = np.arange(istart[ii], iend[ii] + 1)

        # perform blocking of selected block
        (
            rho_b[ii],
            c11_b[ii],
            c13_b[ii],
            c33_b[ii],
            c44_b[ii],
            c66_b[ii],
        ) = backus_average_cij(rho[jj], c11[jj], c13[jj], c33[jj], c44[jj], c66[jj])

    # convert back to Thomsen parameters
    vp_b, vs_b, del_b, eps_b, gam_b = conversions.conv_vti_cij2_thoms(
        rho_b, c11_b, c13_b, c33_b, c44_b, c66_b
    )
    
    # set vs back to zero
    vs_b[vs_b == small_value] = 0

    # output depth
    zz_b = zz[iend]

    # add top layer (elastic values not used)
    zz_b = np.insert(zz_b, 0, zz[0])
    vp_b = np.insert(vp_b, 0, vp_b[0])
    vs_b = np.insert(vs_b, 0, vs_b[0])
    rho_b = np.insert(rho_b, 0, rho_b[0])
    del_b = np.insert(del_b, 0, del_b[0])
    eps_b = np.insert(eps_b, 0, eps_b[0])
    gam_b = np.insert(gam_b, 0, gam_b[0])

    # QC plot
    if qc_plot:

        # collect data in dictionaries
        zz, vp, vs, rho, delta, epsilon, gamma
        model_input = {
            "depth": zz,
            "vp": vp,
            "vs": vs,
            "rhob": rho,
            "delta": delta,
            "epsilon": epsilon,
            "gamma": gamma,
        }
        model_backus = {
            "depth": zz_b,
            "vp": vp_b,
            "vs": vs_b,
            "rhob": rho_b,
            "delta": del_b,
            "epsilon": eps_b,
            "gamma": gam_b,
        }

        # qc plot
        backus_qc_plot(model_input, model_backus)

    return zz_b, vp_b, vs_b, rho_b, del_b, eps_b, gam_b


def backus_qc_plot(model_input, model_backus):

    # define log types
    log_types = list(model_input.keys())[1:]
    ncol = len(log_types)

    # - make figure
    fig = plt.figure("Backus QC")
    fig.clf()

    # - plot original and blocked logs
    icol = 1
    for type_s in log_types:

        plt.subplot(1, ncol, icol)
        plt.cla()
        icol = icol + 1

        plt.plot(model_input[type_s], model_input["depth"], "b-")
        plt.step(model_backus[type_s], model_backus["depth"], "r-", where="post")

        plt.grid()
        plt.title(type_s)
        plt.ylabel("DEPTH (m)")
        plt.gca().invert_yaxis()



import numpy as np


# turn off some warnings


# Alexey Stovas and Bjorn Ursin
# "Reflection and transmission responses of layered transversely isotropic viscoelastic media"
# Geophysical Prospecting, 2003, 51, 447–477


def conv_vti_thoms2cij(rho, vp, vs, delT=None, epsT=None, gamT=None):
    if delT is None:
        delT = np.zeros(np.shape(rho))
    if epsT is None:
        epsT = np.zeros(np.shape(rho))
    if gamT is None:
        gamT = np.zeros(np.shape(rho))

    c33 = rho * vp ** 2
    c44 = rho * vs ** 2
    c11 = c33 * (2 * epsT + 1)
    c66 = c44 * (2 * gamT + 1)
    c13 = np.sqrt(2 * c33 * (c33 - c44) * delT + (c33 - c44) ** 2) - c44

    return c11, c13, c33, c44, c66


def conv_vti_cij2_thoms(rho, c11, c13, c33, c44, c66):
    vp = np.sqrt(c33 / rho)  # vertical P-wave velocity (m/s)
    vs = np.sqrt(c44 / rho)  # vertical S-wave velocity (m/s)
    delT = ((c13 + c44) ** 2 - (c33 - c44) ** 2) / (2 * c33 * (c33 - c44))
    epsT = (c11 - c33) / (2 * c33)
    gamT = (c66 - c44) / (2 * c44)

    return vp, vs, delT, epsT, gamT


def get_greek_param_vti(rho, c11, c13, c33, c44):

    alpha0sq = c33 / rho
    beta0sq = c44 / rho
    sigma0 = 1 - (c44 / c33)

    del0 = (c13 - c33 + 2 * c44) / c33
    eta0 = (c11 * c33 - (c13 + 2 * c44) ** 2) / (2 * c33 ** 2)

    return alpha0sq, beta0sq, sigma0, del0, eta0


def get_phase_vel(th, vp0, vs0, delT, epsT):

    # if input angle is 2D, convert elastic arrays to 2D
    if th.ndim == 2:

        # convert input 1D arrays to 2D
        nn = len(vp0)
        nangles = th.shape[1]
        vp0 = np.tile(vp0.reshape((nn, 1)), (1, nangles))
        vs0 = np.tile(vs0.reshape((nn, 1)), (1, nangles))
        delT = np.tile(delT.reshape((nn, 1)), (1, nangles))
        epsT = np.tile(epsT.reshape((nn, 1)), (1, nangles))

    f = 1 - (vs0 / vp0) ** 2
    temp = (
        1
        + (4 / f)
        * np.sin(th) ** 2
        * (2 * delT * np.cos(th) ** 2 - epsT * np.cos(2 * th))
        + 4 * (epsT / f) ** 2 * np.sin(th) ** 4
    )
    temp = 1 + epsT * np.sin(th) ** 2 - (f / 2) + (f / 2) * np.sqrt(temp)
    vp_ph = vp0 * np.sqrt(temp)

    return vp_ph


def get_layer_qd(p, alpha0sq, beta0sq, sigma0, del0, eta0):  # , opt)

    # ensure all input is treated as complex numbers
    p = p.astype(complex)
    alpha0sq = alpha0sq.astype(complex)
    beta0sq = beta0sq.astype(complex)
    sigma0 = sigma0.astype(complex)
    del0 = del0.astype(complex)
    eta0 = del0.astype(complex)

    # if input slowness is 2D, convert elastic arrays to 2D
    if p.ndim == 2:

        # convert input 1D arrays to 2D
        nn = len(alpha0sq)
        nslow = p.shape[1]
        alpha0sq, beta0sq, sigma0, del0, eta0
        alpha0sq = np.tile(alpha0sq.reshape((nn, 1)), (1, nslow))
        beta0sq = np.tile(beta0sq.reshape((nn, 1)), (1, nslow))
        sigma0 = np.tile(sigma0.reshape((nn, 1)), (1, nslow))
        del0 = np.tile(del0.reshape((nn, 1)), (1, nslow))
        eta0 = np.tile(eta0.reshape((nn, 1)), (1, nslow))

    # q parameters
    R1 = 2 * (1 - p ** 2 * beta0sq) * (del0 + 2 * p ** 2 * alpha0sq * eta0) ** 2
    R2 = (
        sigma0
        + 2 * p ** 2 * beta0sq * del0
        - 2 * p ** 2 * alpha0sq * (1 - 2 * p ** 2 * beta0sq) * eta0
    )
    R = R1 * (R2 + np.sqrt(R2 ** 2 + 2 * p ** 2 * beta0sq * R1)) ** -1

    # % if isequal(opt.type,'weak')
    # %     R=0;
    # % end

    Sa = 2 * del0 + 2 * p ** 2 * alpha0sq * eta0 + R
    Sb = 2 * (1 - p ** 2 * beta0sq) * (alpha0sq / beta0sq) * eta0 - R

    # % if isequal(opt.type,'iso')
    # %     Sa=0; Sb=0;
    # % end

    qa0sq = (1 / alpha0sq) - p ** 2
    qb0sq = (1 / beta0sq) - p ** 2

    tempa = qa0sq - p ** 2 * Sa
    tempb = qb0sq - p ** 2 * Sb
    qa = np.sqrt(tempa)
    qb = np.sqrt(tempb)

    # % qa=-qa;%real(qa)+1i*imag(qa);
    # % qb=-qb;

    # d parameters
    d2 = np.sqrt((sigma0 + del0) / (sigma0 + Sa))
    d3 = 2 * beta0sq * (sigma0 + 0.5 * (Sa + del0)) / (sigma0 + del0)
    d4 = np.sqrt(
        (sigma0 - p ** 2 * beta0sq * (sigma0 + Sb))
        / ((1 - p ** 2 * beta0sq * (1 + Sb)) * (sigma0 + del0))
    )
    d5 = (sigma0 - 2 * p ** 2 * beta0sq * (sigma0 + 0.5 * (Sb + del0))) / (
        sigma0 + del0
    )
    d1 = (p ** 2 * d3 + d5) ** -0.5

    return d1, d2, d3, d4, d5, qa, qb


def get_interface_rt_old(
    p,
    rho_1,
    qa_1,
    qb_1,
    d1_1,
    d2_1,
    d3_1,
    d4_1,
    d5_1,
    rho_2,
    qa_2,
    qb_2,
    d1_2,
    d2_2,
    d3_2,
    d4_2,
    d5_2,
):

    # make top and bottom arrays
    # qa_1, qb_1, d1_1, d2_1, d3_1, d4_1, d5_1 = qa[:-1,:], qb[:-1,:], d1[:-1,:], d2[:-1,:], d3[:-1,:], d4[:-1,:], d5[:-1,:]
    # qa_2, qb_2, d1_2, d2_2, d3_2, d4_2, d5_2 = qa[1:,:], qb[1:,:], d1[1:,:], d2[1:,:], d3[1:,:], d4[1:,:], d5[1:,:]

    # get density ratio, and convert to 2D matrix
    zeta = (rho_2 / rho_1) ** 0.5
    nslow = p.shape[1]
    zeta = np.tile(zeta.reshape((len(zeta), 1)), (1, nslow))

    # Final T and R Matrices
    temp = ((d2_1 * d4_1) ** 2 * qa_1 * qb_1 * (p ** 2 * d3_1 * zeta ** -1 + d5_2 * zeta) ** 2 \
        + (d2_2 * d4_2) ** 2 * qa_2 * qb_2 * (p ** 2 * d3_2 * zeta + d5_1 * zeta ** -1) ** 2 \
        + p ** 2 * (d2_1 * d2_2 * d4_1 * d4_2) ** 2 * qa_1 * qa_2 * qb_1 * qb_2 * (d3_1 * zeta ** -1 - d3_2 * zeta) ** 2 \
        + p ** 2 * (d5_1 * zeta ** -1 - d5_2 * zeta) ** 2)
        
    F = ((d2_2 * d4_1) ** 2 * qa_2 * qb_1 \
        + (d2_1 * d4_2) ** 2 * qa_1 * qb_2 \
        + (d1_1 * d1_2) ** 2 * temp)
        
    # reflection 

    # r_DPP
    temp = ((d2_1 * d4_1) ** 2 * qa_1 * qb_1 * (p ** 2 * d3_1 * zeta ** -1 + d5_2 * zeta) ** 2 \
        - (d2_2 * d4_2) ** 2 * qa_2 * qb_2 * (p ** 2 * d3_2 * zeta + d5_1 * zeta ** -1) ** 2 \
        + p ** 2 * (d2_1 * d2_2 * d4_1 * d4_2) ** 2 * qa_1 * qa_2 * qb_1 * qb_2 * (d3_1 * zeta ** -1 - d3_2 * zeta) ** 2 \
        - p ** 2 * (d5_1 * zeta ** -1 - d5_2 * zeta) ** 2)    
    r_DPP = (1 / F) * ((d2_2 * d4_1) ** 2 * qa_2 * qb_1 \
        - (d2_1 * d4_2) ** 2 * qa_1 * qb_2 \
        - (d1_1 * d1_2) ** 2 * temp)
    r_DPP = -r_DPP  # Don't know why I must change the polarity (??)
    
    # r_DSS
    temp = ((d2_1 * d4_1) ** 2 * qa_1 * qb_1 * (p ** 2 * d3_1 * zeta ** -1 + d5_2 * zeta) ** 2 \
        - (d2_2 * d4_2) ** 2 * qa_2 * qb_2 * (p ** 2 * d3_2 * zeta + d5_1 * zeta ** -1) ** 2 \
        + p ** 2 * (d2_1 * d2_2 * d4_1 * d4_2) ** 2 * qa_1 * qa_2 * qb_1 * qb_2 * (d3_1 * zeta ** -1 - d3_2 * zeta) ** 2 \
        - p ** 2 * (d5_1 * zeta ** -1 - d5_2 * zeta) ** 2)    
    r_DSS = (1 / F) * ((d2_2 * d4_1) ** 2 * qa_2 * qb_1 \
        - (d2_1 * d4_2) ** 2 * qa_1 * qb_2 \
        + (d1_1 * d1_2) ** 2 * temp)
    r_DSS = -r_DSS  # Don't know why I must change the polarity (??)    
    
    # r_DPS
    temp = (d2_2 * d4_2) ** 2 * qa_2 * qb_2 * (p ** 2 * d3_2 * zeta + d5_1 * zeta ** -1) * (d3_1 * zeta ** -1 - d3_2 * zeta) \
        + (p ** 2 * d3_1 * zeta ** -1 + d5_2 * zeta) * (d5_1 * zeta ** -1 - d5_2 * zeta)
    r_DPS = ((2 * p / F) * (d1_1 * d1_2) ** 2 * d2_1 * d4_1 * (qa_1 * qb_1) ** 0.5 * temp)
    r_DSP = r_DPS    
    
    
    

    # a = (d2_2 * d4_1) ** 2 * qa_2 * qb_1
    # b = (d2_1 * d4_2) ** 2 * qa_1 * qb_2
    # c = (d1_1 * d1_2) ** 2 * temp

    qa_1_sqroot = qa_1 ** 0.5
    qb_1_sqroot = qb_1 ** 0.5
    # qa_2_sqroot = qa_2 ** .5
    # qb_2_sqroot = qb_2 ** .5

    temp = (d2_2 * d4_2) ** 2 * qa_2 * qb_2 * (
        p ** 2 * d3_2 * zeta + d5_1 * zeta ** -1
    ) * (d3_1 * zeta ** -1 - d3_2 * zeta) + (
        p ** 2 * d3_1 * zeta ** -1 + d5_2 * zeta
    ) * (
        d5_1 * zeta ** -1 - d5_2 * zeta
    )
    r_DPS = (
        (2 * p / F)
        * (d1_1 * d1_2) ** 2
        * d2_1
        * d4_1
        * (qa_1_sqroot * qb_1_sqroot)
        * temp
    )
    r_DSP = r_DPS
    
    # t_DPP
    temp = d4_2 ** 2 * qb_2 * (p ** 2 * d3_2 * zeta + d5_1 * zeta ** -1) \
        + d4_1 ** 2 * qb_1 * (p ** 2 * d3_1 * zeta ** -1 + d5_2 * zeta)        
    t_DPP = (2 / F) * d1_1 * d1_2 * d2_1 * d2_2 * (qa_1 * qa_2) ** 0.5 * temp    

    out = {"r_DPP": r_DPP, "r_DPS": r_DPS, "r_DSP": r_DSP, "r_DSS": r_DSS, "t_DPP": t_DPP}
    return out

def get_interface_rt(
    p,
    rho_1,
    qa_1,
    qb_1,
    d1_1,
    d2_1,
    d3_1,
    d4_1,
    d5_1,
    rho_2,
    qa_2,
    qb_2,
    d1_2,
    d2_2,
    d3_2,
    d4_2,
    d5_2,
):
    
    # get density ratio, and convert to 2D matrix
    zeta = (rho_2 / rho_1) ** 0.5
    nslow = p.shape[1]
    zeta = np.tile(zeta.reshape((len(zeta), 1)), (1, nslow))    

    
    # C matrix
    c11 = (d2_2/d2_1) * (qa_2/qa_1)**0.5 * (p**2 * d3_2 * zeta + d5_1 * zeta**-1)
    c12 = (d2_1 * d4_2)**-1 * (qa_1 * qb_2)**-0.5 * p * (d5_1 * zeta**-1 - d5_2*zeta)
    c21 = (d2_2 * d4_1) * (qa_2 * qb_1)**0.5 * p * (d3_1 * zeta**-1 - d3_2*zeta)
    c22 = (d4_1/d4_2) * (qb_1/qb_2)**0.5 * (p**2 * d3_1 * zeta**-1 + d5_2 * zeta)
    
    c11 = d1_1 * d1_2 * c11
    c12 = d1_1 * d1_2 * c12
    c21 = d1_1 * d1_2 * c21
    c22 = d1_1 * d1_2 * c22
    
    det_C = c11*c22 - c21*c12
    
    # D matrix
    d11 = (1/det_C) * c22
    d12 = (1/det_C) * -c21
    d21 = (1/det_C) * -c12
    d22 = (1/det_C) * c11
    
    # make R_D matrix
    temp = (det_C**2 + 1 + c11**2 + c12**2 + c21**2 + c22**2)**-1   
    r_DPP = temp * (det_C**2 - 1 + c11**2 + c12**2 - c21**2 - c22**2)
    r_DPP = -r_DPP # must change sign, don't know why
    r_DSS = temp * (det_C**2 - 1 - c11**2 - c12**2 + c21**2 + c22**2)    
    r_DPS = temp * 2 * (c11*c21 + c12*c22)
    r_DSP = temp * 2 * (c11*c21 + c12*c22)
    
    # make T_D matrix
    t_DPP = 2 * temp * (c11 + c22 * det_C)
    
    # return output
    out = {"r_DPP": r_DPP, "r_DPS": r_DPS, "r_DSP": r_DSP, "r_DSS": r_DSS, "t_DPP": t_DPP}
    return out


def get_tr_rc_vti(vp, vs, rho, delT, epsT, gamT, th_deg):
    
    # get transmission and reflection coefficients for layered vti media

    # turn off some warnings
    np.seterr(divide="ignore", invalid="ignore")

    # set anisotorpy to zero if empty
    if len(delT) == 0:
        delT = 0 * vp
    if len(epsT) == 0:
        epsT = 0 * vp
    if len(gamT) == 0:
        gamT = 0 * vp

    # add small number to vs to avoid division by zero
    vs = vs + 1e-12

    # if input angles are 1D array, make 2D array
    nlayers = len(vp)
    if th_deg.ndim == 1:
        th_deg = np.tile(th_deg, (nlayers, 1))

    # convert angles to radians and ensure complex type
    th_rad = (np.pi / 180) * th_deg

    # get phase velocity as function of angle
    vp_ph = get_phase_vel(th_rad, vp, vs, delT, epsT)

    # slowness
    pp = np.sin(th_rad) / vp_ph

    # get Thomsen parameters
    c11, c13, c33, c44, c66 = conv_vti_thoms2cij(
        rho, vp, vs, delT=delT, epsT=epsT, gamT=gamT
    )

    # get additional parameters
    alpha0sq, beta0sq, sigma0, del0, eta0 = get_greek_param_vti(rho, c11, c13, c33, c44)

    # get slowness dependent parameters    
    d1_1, d2_1, d3_1, d4_1, d5_1, qa_1, qb_1 = get_layer_qd(
        pp[:-1, :], alpha0sq[:-1], beta0sq[:-1], sigma0[:-1], del0[:-1], eta0[:-1]
    )
    d1_2, d2_2, d3_2, d4_2, d5_2, qa_2, qb_2 = get_layer_qd(
        pp[:-1, :], alpha0sq[1:], beta0sq[1:], sigma0[1:], del0[1:], eta0[1:]
    )

    # get interface reflection and transmission coefficients
    ref = get_interface_rt(
        pp[:-1, :],
        rho[:-1],
        qa_1,
        qb_1,
        d1_1,
        d2_1,
        d3_1,
        d4_1,
        d5_1,
        rho[1:],
        qa_2,
        qb_2,
        d1_2,
        d2_2,
        d3_2,
        d4_2,
        d5_2,
    )
    
    # add NaN base layer (for consistent sizes)    
    vv = np.nan * np.ones((1, th_deg.shape[1]))
    for key in ref:            
        ref[key] = np.concatenate([ref[key], vv])       
    
    # turn back on warnings
    np.seterr(all="warn")    
    
    # return reflection/transmission coefficients
    return ref


def get_rpp_vti(vp, vs, rho, delT, epsT, gamT, th_deg, apply_tr_loss=False):
    
    # get PP reflection coefficients for layered vti media
    
    
    # get all reflection and transmission coefficients
    ref = get_tr_rc_vti(vp, vs, rho, delT, epsT, gamT, th_deg)

    
    # select downgoing PP ref. coef
    RC = ref["r_DPP"]
    
    # apply transmission loss
    if apply_tr_loss:
        
        print('WARNING: Transmission loss is not yet implemented for VTI media.')
        
        #R = ref["r_DPP"] # reflection coefficient
        #T = ref["t_DPP"] # transmission coefficient
        

    # add bottom layer (not used)
    #RC = np.concatenate([RC, np.nan * np.ones((1, RC.shape[1]))])

    # turn back on warnings
    np.seterr(all="warn")

    # return result
    return RC

def get_ruger_par(vp, vs, rho, delT, epsT, gamT, p):
                  
    # get elastic parameters
    c11, c13, c33, c44, c66 = conv_vti_thoms2cij(rho, vp, vs, delT=delT, epsT=epsT, gamT=gamT) 
    
    # rename elastic parameters according to convention in Rüger
    c55 = c44
    
    # make a parameters
    a11, a13, a33, a55, a66 = c11/rho, c13/rho, c33/rho, c55/rho, c66/rho
    
    # make K parameters
    K1 = (rho/c33) + (rho/c55) - ((c11/c55) + (c55/c33) - ((c13 + c55)**2/(c33*c55)) ) * p**2
    K2 = (c11/c33)*p**2 - (rho/c33)
    K3 = p**2 - (rho/c55)
    
    # make q parameters
    qa = (1/np.sqrt(2)) * np.sqrt(K1 - np.sqrt(K1**2 - 4*K2*K3))
    qb = (1/np.sqrt(2)) * np.sqrt(K1 + np.sqrt(K1**2 - 4*K2*K3))
    
    # make l and m parameters        
    temp = a11*p**2 + a55*qa**2 - 1 + a33*qa**2 + a55*p**2 - 1
    la = ((a33*qa**2 + a55*p**2 - 1)/temp)**0.5        
    lb = ((a11*p**2 + a55*qb**2 - 1)/temp)**0.5
    ma = ((a11*p**2 + a55*qa**2 - 1)/temp)**0.5    
    mb = ((a11*qb**2 + a55*p**2 - 1)/temp)**0.5
        
    # return layer parameters    
    return a11, a13, a33, a55, a66, qa, qb, la, lb, ma, mb    
    

def get_rpp_vti_ruger(vp, vs, rho, delT, epsT, gamT, th_deg, apply_tr_loss=False):
        
    # Rüger 
    
    # get size of input
    nlayer = len(vp)
    nth = th_deg.shape[1]
    
    # initiate output
    r_dpp = np.zeros(th_deg.shape)
        
    # convert angles to radians 
    th_rad = (np.pi / 180) * th_deg

    # get phase velocity as function of angle
    vp_ph = get_phase_vel(th_rad, vp, vs, delT, epsT).astype(complex)

    # slowness
    pp = np.sin(th_rad) / vp_ph 
    
    print(pp)
    

    for ii in np.arange(nlayer-1): # loop over layers   
        for jj in np.arange(nth):   
            
            # single value slowness
            p = pp[ii,jj]
    
            # get values for upper layer
            a11_1, a13_1, a33_1, a55_1, a66_1, qa_1, qb_1, la_1, lb_1, ma_1, mb_1 = get_ruger_par(
                vp[ii], vs[ii], rho[ii], delT[ii], epsT[ii], gamT[ii], pp[ii,jj])
            
            # get values for lower layer
            a11_2, a13_2, a33_2, a55_2, a66_2, qa_2, qb_2, la_2, lb_2, ma_2, mb_2 = get_ruger_par(
                vp[ii+1], vs[ii+1], rho[ii+1], delT[ii+1], epsT[ii+1], gamT[ii+1], pp[ii,jj])
                            
            # elements of M matrix
            m11 = la_1
            m12 = mb_1
            m13 = -la_2
            m14 = -mb_2
            m31 = ma_1
            m32 = -lb_1
            m33 = ma_2
            m34 = -lb_2
            m21 = p * la_1 * a13_1 + qa_1 * ma_1 * a33_1
            m22 = p * mb_1 * a13_1 - qb_1 * lb_1 * a33_1
            m23 = -(p * la_2 * a13_2 + qa_2 * ma_2 * a33_2)
            m24 = -(p * mb_2 * a13_2 - qb_2 * lb_2 * a33_2)
            m41 = a55_1 * (qa_1 * la_1 + p * ma_1)
            m42 = a55_1 * (qb_1 * mb_1 - p * lb_1)
            m43 = a55_2 * (qa_2 * la_2 + p * ma_2)
            m44 = a55_2 * (qb_2 * mb_2 - p * lb_2)    
            
            # make linear system
            M = np.array([[m11, m12, m13, m14], [m21, m22, m23, m24], [m31, m32, m33, m34], [m41, m42, m43, m44]]).astype(complex)
            b = np.array([-m11, -m21, m31, m41]).astype(complex)
            
            print(M)
            print(b)
            
            # solve linear system
            x = np.linalg.solve(M, b)
            print('x')
            print(x)
            
            # assign solution
            r_dpp[ii,jj] = x[0]
            
            
            
    return r_dpp
    
    
    
    
    
    
    
    
    
    

























import numpy as np
from scipy.interpolate import interp1d
import matplotlib.pyplot as plt


def get_iso(tv, zv, xv, direction=1, nterm=4, tv_output=[]):

    # get moveout curves for isotropic avo
    # Based on Sun, Wang and Martinez (SEG 2002)
    #
    # ported from Matlab
    #
    # tv : times. 1D array of length n
    # zv: corresponding depth of tv0
    # xv : offsets. 1D array of length m
    # nterm: number of terms to include
    # direction: 1 or -1

    # get some parameters
    c1, c2, c3, c4 = get_par_iso(tv, zv)

    # select number of terms
    if nterm < 4:
        c4 = 0 * c4
    if nterm < 3:
        c3 = 0 * c3

    # change shapes
    xv = xv.reshape((1, len(xv)))
    c1 = c1.reshape((len(c1), 1))
    c2 = c2.reshape((len(c2), 1))
    c3 = c3.reshape((len(c3), 1))
    c4 = c4.reshape((len(c4), 1))

    # calculate output
    tmp = np.tile(c1, len(xv)) + direction * (
        np.matmul(c2, xv ** 2) + np.matmul(c3, xv ** 4) + np.matmul(c4, xv ** 6)
    )
    TX = np.sqrt(np.maximum(0, tmp))

    # calculate moveout curves at specific outptut times
    if len(tv_output) > 0:
        for ix in np.arange(len(xv)):
            TX_new = np.zeros((len(tv_output), len(xv)))
            ff = interp1d(
                tv,
                TX[:, ix],
                kind="linear",
                bounds_error=False,
                fill_value="extrapolate",
            )
            TX_new[:, ix] = ff(tv_output)
        TX = TX_new

    # return movout matrix
    return TX


def get_par_iso(tv, zv):

    # Sun, Wang and Martinez (SEG 2002)
    #
    # This function will calculate a set of coefficients that will be used to
    # get the two-way traveltime after a higher order NMO correction. The
    # coefficients will be used as follows:
    # T = c_1 + c_2*x^2 + c_3*x^4 + c_4*x^6

    # get velocity (m/s)
    dz = np.diff(zv)
    dt = np.diff(tv)
    vv = np.array([2 * (dz / dt)])

    # calculate some parameters
    a1 = 2 * np.cumsum(dz * vv ** -1)
    a2 = 2 * np.cumsum(dz * vv ** 1)
    a3 = 2 * np.cumsum(dz * vv ** 3)
    a4 = 2 * np.cumsum(dz * vv ** 5)

    # calculate output
    c1 = a1 ** 2
    c2 = a1 / a2
    c3 = (a1 ** 2 - a1 * a3) / (4 * a2 ** 4)
    c4 = (2 * a1 * a3 ** 2 - a1 * a2 * a4 - a2 ** 2 * a3) / (8 * a2 ** 7)

    # insert zero at start of each array and ensure 2D output
    c1 = np.insert(c1, 0, 0)
    c2 = np.insert(c2, 0, 0)
    c3 = np.insert(c3, 0, 0)
    c4 = np.insert(c4, 0, 0)

    # return result
    return c1, c2, c3, c4


def get_vti(tv, zv, xv, depth_model, vs_model, delta_model, epsilon_model, direction=1):

    # function [traces TX]=nmo_vti_applyCorr(traces0,xv0,tv0,zv0,t_top0,vs0,delT0,epsT0,opt)
    # % traces0 : input traces. array with n rows and m columns
    # % xv0 : offsets (m). 1D array of length m
    # % tv0 : times (ms). 1D array of length n
    # % zv0 : corresponding depth (m) of tv0 n
    # % opt.dir : 1 - normal NMO, -1 - reverse NMO
    # %
    # % Then a few parameters defining k layers
    # % t_top (ms): length k+1
    # % vs (m/s), delT, epsT, gamT : length k
    # % (vp is derived from tv0 and zv0 - model might have been stretched prior to NMO correction)
    # %
    # % Ursin and Stovas (2006)

    # resample input layers to time vector of traces
    ff = interp1d(zv, tv, kind="linear", bounds_error=False, fill_value="extrapolate")
    time_model = ff(depth_model)
    vs = resample_layer_model(time_model, vs_model, tv)
    delta = resample_layer_model(time_model, delta_model, tv)
    epsilon = resample_layer_model(time_model, epsilon_model, tv)

    # derive vp from time-depth  input
    dt = np.diff(tv, axis=0)
    dz = np.diff(zv, axis=0)
    vp = 2 * dz / dt  # vp=[vp; vp(end)];
    vp = np.insert(vp, 0, vp[0])
    # vp = vp.reshape((1,len(vp)))

    #  get parameters for continued-fraction formula
    vnmo_sq, c1, c2, c3, c4, B = get_par_vti(tv, vp, vs, delta, epsilon)

    # loop through traces and create time table
    TX = np.zeros((len(tv), len(xv)))  # initiate output
    for ix in np.arange(len(xv)):

        # current offset
        xx = xv[ix]

        # get move-out time (squared)
        if 1:  # % Taylor expansion
            temp = 1 + (B * xx ** 2) / (tv ** 2 * vnmo_sq)
            tx_sq = (1 / vnmo_sq) * xx ** 2 + (c2 * xx ** 4) / temp
        else:  # % continued fraction
            tx_sq = c1 * xx ** 2 + c2 * xx ** 4 + c3 * xx ** 6 + c4 * xx ** 8

        tx_sq = tv ** 2 + direction * tx_sq
        tx = np.sqrt(np.maximum(0, tx_sq))

        # assign to output
        TX[:, ix] = tx

    # return output
    return TX


def plot_tx(TX, tv, xv):

    # make figure
    fig = plt.figure("NMO QC2")
    fig.clf()
    nsub = 1
    isub = 1

    plt.subplot(1, nsub, isub)
    isub = isub + 1
    im = plt.imshow(
        TX, aspect="auto"
    )  # ", vmin =, vmax = np.nanmax(self.time[-1,:]), cmap = cmap, aspect='auto', interpolation='none')
    plt.grid()
    # plt.title('time (s)')
    fig.colorbar(im, ax=plt.gca())


def apply(
    traces_in,
    tv,
    zv,
    xv,
    depth_model=[],
    vs_model=[],
    delta_model=[],
    epsilon_model=[],
    nmo_type="isotropic",
    tv_output=[],
    direction=1,
    nterm=4,
):

    # get movout curves
    TX = get(
        tv,
        zv,
        xv,
        depth_model=depth_model,
        vs_model=vs_model,
        delta_model=delta_model,
        epsilon_model=epsilon_model,
        nmo_type=nmo_type,
        tv_output=tv_output,
        direction=direction,
        nterm=nterm,
    )

    # return output traces
    return apply_tx(traces_in, tv, TX)


def resample_tx(TX, tv_out):

    nx = TX.shape[1]
    nt_out = len(tv_out)

    TX_out = np.zeros((nt_out, nx))
    for ix in np.arange(nx):
        ff = interp1d(
            TX[:, 0],
            TX[:, ix],
            kind="linear",
            bounds_error=False,
            fill_value="extrapolate",
        )
        TX_out[:, ix] = ff(tv_out)
    return TX_out


def apply_tx(traces_in, tv, TX):

    # initiate output
    traces_out = 0 * traces_in

    # loop through traces and apply
    for ix in np.arange(traces_in.shape[1]):

        # interpolate
        ff = interp1d(
            tv,
            traces_in[:, ix],
            kind="linear",
            bounds_error=False,
            fill_value="extrapolate",
        )
        traces_out[:, ix] = ff(TX[:, ix])

    # return output
    return traces_out


def get(
    tv,
    zv,
    xv,
    depth_model=[],
    vs_model=[],
    delta_model=[],
    epsilon_model=[],
    nmo_type="isotropic",
    tv_output=[],
    direction=1,
    nterm=4,
):

    # turn off some warnings
    np.seterr(divide="ignore", invalid="ignore", under="ignore")

    # get moveout curves
    if nmo_type == "isotropic":

        TX = get_iso(tv, zv, xv, direction=direction, nterm=nterm)

    elif nmo_type == "vti":

        TX = get_vti(
            tv,
            zv,
            xv,
            depth_model,
            vs_model,
            delta_model,
            epsilon_model,
            direction=direction,
        )

    # calculate moveout curves at specific outptut times
    if len(tv_output) > 0:
        TX_new = np.zeros((len(tv_output), len(xv)))
        for ix in np.arange(len(xv)):
            ff = interp1d(
                tv,
                TX[:, ix],
                kind="linear",
                bounds_error=False,
                fill_value="extrapolate",
            )
            TX_new[:, ix] = ff(tv_output)
        TX = TX_new

    # QC plot
    if 0:
        plot_tx(TX, tv, xv)

    # turn back on warnings
    np.seterr(all="warn")

    # return output
    return TX


def resample_layer_model(time_model, val_model, tv_out):

    ff = interp1d(
        time_model, val_model, kind="next", bounds_error=False, fill_value="extrapolate"
    )
    return ff(tv_out)


def get_par_vti(t0, vp, vs, delT, epsT):

    # define some parameters
    v0 = vp
    gam0 = vp / vs
    sig = gam0 ** 2 * (epsT - delT)

    b1 = gam0 ** 2 - 1
    b2 = 1 + (2 * delT * gam0 ** 2) / b1
    b3 = (4 * sig) / (b1 * gam0 ** 2)

    a0 = 1 + 2 * delT
    a1 = ((2 * sig) / (gam0 ** 2)) * b1
    a2 = -b3 * (delT - sig) * b2
    temp = sig * b2 - 2 * gam0 ** 2 * (delT - sig) ** 2 / b1
    a3 = -(b3 / (gam0 ** 2)) * b2 * temp

    # replace NaN & Inf with zeros (from waterlayer)
    a1[np.isinf(a1)] = 0
    a2[np.isinf(a2)] = 0
    a3[np.isinf(a3)] = 0

    a1[np.isnan(a1)] = 0
    a2[np.isnan(a2)] = 0
    a3[np.isnan(a3)] = 0

    # time step
    dt = t0[1] - t0[0]

    #  mu parameters
    mu0 = np.ones((len(t0), 1))
    mu2 = (dt / t0) * np.cumsum(v0 ** 2 * a0)
    vnmo_sq = mu2
    mu4 = (dt / t0) * np.cumsum(v0 ** 4 * (a0 ** 2 + 4 * a1))
    mu6 = (dt / t0) * np.cumsum(v0 ** 6 * (a0 ** 3 + 4 * a1 * a0 + 8 * a2))
    temp = (
        a0 ** 4
        + (24 / 5) * a1 * a0 ** 2
        + (32 / 5) * a2 * a0
        + (16 / 5) * a1 ** 2
        + (64 / 5) * a3
    )
    mu8 = (dt / t0) * np.cumsum(v0 ** 8 * temp)

    # make S parameters
    S2 = mu4 / mu2 ** 2
    S3 = mu6 / mu4 ** 2
    S4 = mu8 / mu6 ** 2

    # some more parameters
    c2 = (1 - S2) / (4 * t0 ** 2 * vnmo_sq ** 2)
    B = (2 * S2 ** 2 - S2 - S3) / (2 * (S2 - 1))

    c1 = 1 / vnmo_sq
    c3 = (2 * S2 ** 2 - S2 - S3) / (8 * t0 ** 4 * vnmo_sq ** 3)
    temp1 = 24 * (S3 - S2 ** 2) * S2 + 9 * S2 ** 2 - 4 * S3 - 5 * S4
    temp2 = 64 * t0 ** 6 * vnmo_sq ** 4
    c4 = temp1 / temp2

    # return output
    return vnmo_sq, c1, c2, c3, c4, B



import numpy as np
import pandas as pd
from sklearn.linear_model import LinearRegression


def get_intercept_gradient(traces, theta, max_angle=25):

    # Returns intercept and gradient stacks of AVA gather

    # initiate output
    nt, nth = traces.shape
    intercept_stack = np.zeros(nt)
    gradient_stack = np.zeros(nt)

    # remove data above max angle
    traces = traces[:, theta <= max_angle]
    theta = theta[theta <= max_angle]

    # return if empty
    if len(traces) == 0:
        return intercept_stack, gradient_stack

    # create scaled angle vector
    theta_scaled = np.sin(theta * (np.pi / 180)) ** 2

    # loop through all time samples and derive intercept/gradient from linear fit
    ii = 0
    for row in traces:
        xx = theta_scaled.reshape((-1, 1))
        yy = row
        model = LinearRegression().fit(xx, yy)
        intercept_stack[ii] = model.intercept_
        gradient_stack[ii] = model.coef_
        ii = ii + 1

    # return intercept/gradient
    return intercept_stack, gradient_stack


def get_eei_stack(intercept_stack, gradient_stack, rotation_angle_deg):

    # get angle in radians
    th_rad = rotation_angle_deg * (np.pi / 180)

    # create rotated stack
    eei_stack = np.cos(th_rad) * intercept_stack + np.sin(th_rad) * gradient_stack

    # return ei stack
    return eei_stack


def get_angle_stack(traces, theta, angle_min=0, angle_max=30):

    # get angle stack
    angle_stack = traces[:, (theta >= angle_min) & (theta <= angle_max)].mean(axis=1)

    # return ei stack
    return angle_stack


def get_avo_stacks(traces, theta, avo_stack_opt):

    # returns a set of stacks in a pandas dataframe

    # initiate output
    avo_stacks = pd.DataFrame()

    # get size of input
    nt, nth = traces.shape

    # create near and far stacks
    avo_stacks["NEAR"] = get_angle_stack(
        traces,
        theta,
        angle_min=avo_stack_opt["near_theta_min"],
        angle_max=avo_stack_opt["near_theta_max"],
    )
    avo_stacks["FAR"] = get_angle_stack(
        traces,
        theta,
        angle_min=avo_stack_opt["far_theta_min"],
        angle_max=avo_stack_opt["far_theta_max"],
    )
    avo_stacks["FULL"] = get_angle_stack(
        traces,
        theta,
        angle_min=avo_stack_opt["full_theta_min"],
        angle_max=avo_stack_opt["full_theta_max"],
    )
    # avo_stacks['NEAR'] = traces[:, (theta >= avo_stack_opt['near_theta_min']) & (theta <= avo_stack_opt['near_theta_max'])].mean(axis = 1)
    # avo_stacks['FAR'] = traces[:, (theta >= avo_stack_opt['far_theta_min']) & (theta <= avo_stack_opt['far_theta_max'])].mean(axis = 1)

    # create intercept/gradient stacks
    intercept_stack, gradient_stack = get_intercept_gradient(
        traces, theta, max_angle=avo_stack_opt["max_angle_int_grad"]
    )
    avo_stacks["INT"] = intercept_stack
    avo_stacks["GRAD"] = gradient_stack

    # get EEI stacks
    avo_stacks["LITHO"] = get_eei_stack(
        intercept_stack, gradient_stack, avo_stack_opt["litho_rot_angle"]
    )
    avo_stacks["FLUID"] = get_eei_stack(
        intercept_stack, gradient_stack, avo_stack_opt["fluid_rot_angle"]
    )
    avo_stacks["VPVS"] = get_eei_stack(
        intercept_stack, gradient_stack, avo_stack_opt["vpvs_rot_angle"]
    )
    avo_stacks["SI"] = get_eei_stack(
        intercept_stack, gradient_stack, avo_stack_opt["si_rot_angle"]
    )

    return avo_stacks



import numpy as np
from seismod1d.model import reflectivity_vti, reflectivity_iso  

def multi_layer(
    reflection_angles,
    vp,
    vs,
    rho,
    delta=[],
    epsilon=[],
    gamma=[],
    apply_tr_loss = False,
    ref_method="zoeppritz_rpp",
):


    """
    Return reflection coefficient matrix for multi layered 1D input model.

    Args:
        reflection_angles (ndarray): Incidence angle (degree): 1D array of length n 
        (same angles for all layers) or m x n array (different incidense angles for each layer)
        vp (ndarray): P-wave velocity (m/s); 1D array length m.
        vs (ndarray):  S-wave velocity (m/s);  1D array length m.
        rho (ndarray): Density (g/cm3); 1D array length m.
        delta (ndarray): Thomsen anisotropic delta; empty (if isotropic) or 1D array length m.
        epsilon (ndarray): Thomsen anisotropic delta; empty (if isotropic) or 1D array length m.
        gamma (ndarray): Thomsen anisotropic delta; empty (if isotropic) or 1D array length m.

    Returns:
        RC (ndarray): Reflection coefficient; 2D array of size m x n. 
        First and last row will be nan's (interface not defined).
    """

    # get reflectivity of multiple layers with different reflection angle for each interface

    # ensure positive vs
    vs[vs == 0] = 1e-12

    # get number of input layers (must be at least 2)
    nlayers = len(vp)

    # if input angles is 1D array, make 2D array (one row per interface)
    if reflection_angles.ndim == 1:

        # make 2D array with N rows (same incidence angle for all layers)
        reflection_angles = np.tile(reflection_angles, (nlayers, 1))

    # get number of unique angles
    nangles = reflection_angles.shape[1]

    # initiate output reflection coefficient 2D Array
    RC = np.ones((nlayers, nangles)) * np.nan

    # Get reflectivity of multiple layers
    if ref_method == "zoeppritz_rpp":  # isotropic

        # get isotropic reflection coefficient       
        RC = reflectivity_iso.get_zoeppritz_rpp(
            vp, vs, rho, 
            reflection_angles, apply_tr_loss = apply_tr_loss
        )
            

    elif ref_method == "vti_exact_rpp":  # anisotropic

        # get anisotropic relectivity coefficients
        RC = reflectivity_vti.get_rpp_vti(
            vp, vs, rho, delta, epsilon, gamma, 
            reflection_angles, apply_tr_loss = apply_tr_loss
        )

        
    elif ref_method == "vti_exact_rpp_diff":  # difference anisotropic vs isotropic

        # get isotropic reflection coefficient       
        RC_iso = reflectivity_iso.get_zoeppritz_rpp(
            vp, vs, rho, 
            reflection_angles, apply_tr_loss = apply_tr_loss
        )

        # get anisotropic relectivity coefficients
        RC_vti = reflectivity_vti.get_rpp_vti(
            vp, vs, rho, delta, epsilon, gamma, 
            reflection_angles, apply_tr_loss = apply_tr_loss
        )

        # difference
        RC = RC_vti - RC_iso

    # - return result
    return RC



import numpy as np

def zoeppritz_rpp(vp1, vs1, rho1, vp2, vs2, rho2, theta1=0):
    """
    Exact Zoeppritz from expression.

    This is useful because we can pass arrays to it, which we can't do to
    scattering_matrix().

    Dvorkin et al. (2014). Seismic Reflections of Rock Properties. Cambridge.

    Returns the complex reflectivity.

    Args:
        vp1 (ndarray): The upper P-wave velocity; float or 1D array length m.
        vs1 (ndarray): The upper S-wave velocity; float or 1D array length m.
        rho1 (ndarray): The upper layer's density; float or 1D array length m.
        vp2 (ndarray): The lower P-wave velocity; float or 1D array length m.
        vs2 (ndarray): The lower S-wave velocity; float or 1D array length m.
        rho2 (ndarray): The lower layer's density; float or 1D array length m.
        theta1 (ndarray): The incidence angle; float or 1D array length n.

    Returns:
        ndarray. The exact Zoeppritz solution for P-P reflectivity at the
            interface. Will be a float (for float inputs and one angle), a
            1 x n array (for float inputs and an array of angles), a 1 x m
            array (for float inputs and one angle), or an n x m array (for
            array inputs and an array of angles).
    """
    theta1 = np.radians(theta1).astype(complex)

    # turn off some warnings
    # np.seterr(divide='ignore', invalid='ignore')

    # convert input 1D arrays to 2D
    nn = len(vp1)
    nangles = theta1.shape[1]
    vp1 = np.tile(vp1.reshape((nn, 1)), (1, nangles))
    vs1 = np.tile(vs1.reshape((nn, 1)), (1, nangles)) + 1e-12
    rho1 = np.tile(rho1.reshape((nn, 1)), (1, nangles))
    vp2 = np.tile(vp2.reshape((nn, 1)), (1, nangles))
    vs2 = np.tile(vs2.reshape((nn, 1)), (1, nangles)) + 1e-12
    rho2 = np.tile(rho2.reshape((nn, 1)), (1, nangles))

    p = np.sin(theta1) / vp1  # Ray parameter
    theta2 = np.arcsin(p * vp2)
    phi1 = np.arcsin(p * vs1)  # Reflected S
    phi2 = np.arcsin(p * vs2)  # Transmitted S

    a = rho2 * (1 - 2 * np.sin(phi2) ** 2.0) - rho1 * (1 - 2 * np.sin(phi1) ** 2.0)
    b = rho2 * (1 - 2 * np.sin(phi2) ** 2.0) + 2 * rho1 * np.sin(phi1) ** 2.0
    c = rho1 * (1 - 2 * np.sin(phi1) ** 2.0) + 2 * rho2 * np.sin(phi2) ** 2.0
    d = 2 * (rho2 * vs2 ** 2 - rho1 * vs1 ** 2)

    E = (b * np.cos(theta1) / vp1) + (c * np.cos(theta2) / vp2)
    F = (b * np.cos(phi1) / vs1) + (c * np.cos(phi2) / vs2)
    G = a - d * np.cos(theta1) / vp1 * np.cos(phi2) / vs2
    H = a - d * np.cos(theta2) / vp2 * np.cos(phi1) / vs1

    D = E * F + G * H * p ** 2

    rpp = (1 / D) * (
        F * (b * (np.cos(theta1) / vp1) - c * (np.cos(theta2) / vp2))
        - H * p ** 2 * (a + d * (np.cos(theta1) / vp1) * (np.cos(phi2) / vs2))
    )

    return np.squeeze(rpp)

        
def get_zoeppritz_rpp(vp, vs, rho, reflection_angles, apply_tr_loss = False):
    
    # get all 
    ref = get_zoeppritz(vp, vs, rho, reflection_angles)
        
    # reflection coefficients        
    RC = ref["r_DPP"]
    
    # apply transmission loss    
    if apply_tr_loss:
        
        rd = ref["r_DPP"]
        td = ref["t_DPP"]
        tu = ref["t_UPP"]
        
        td = np.cumprod(td, axis=0)
        tu = np.cumprod(tu, axis=0)
        
        rd[1:,:] = rd[1:,:]*td[:-1,:]*tu[:-1,:]        
        
        RC = rd        
           
    
    return RC

def get_zoeppritz(vp, vs, rho, reflection_angles):
    
        
    # Amundsen et. al
    
    # turn off some warnings
    np.seterr(divide="ignore", invalid="ignore", under="ignore")    
    
    # get number of input layers
    nlayers = len(vp)    
    
    # if input angles is 1D array, make 2D array (one row per interface)
    if reflection_angles.ndim == 1:

        # make 2D array with N rows (same incidence angle for all layers)
        reflection_angles = np.tile(reflection_angles, (nlayers, 1))    
        
    # add small number to vs to ensure non-zero values
    vs[vs==0] = vs[vs==0] + 1e-12
    
    # convert input 1D model arrays to 2D        
    nangles = reflection_angles.shape[1]
    vp = np.tile(vp.reshape((nlayers, 1)), (1, nangles))
    vs = np.tile(vs.reshape((nlayers, 1)), (1, nangles))
    rho = np.tile(rho.reshape((nlayers, 1)), (1, nangles))
    
    # split model into upper and lower layer values
    vp1 = vp[:-1,:]
    vs1 = vs[:-1,:]
    rho1 = rho[:-1,:]
    vp2 = vp[1:,:]
    vs2 = vs[1:,:]
    rho2 = rho[1:,:]

    # convert reflection angles from degree to radian (ignore last row)
    #th_rad = (np.pi/180)*reflection_angles[:-1,:]   
    th_rad = np.radians(reflection_angles[:-1,:]).astype(complex)
    
    # get slowness
    p = np.sin(th_rad)/vp1#.astype(complex)
        
    mu1 = rho1 * vs1**2
    mu2 = rho2 * vs2**2
    
    qp1 = np.sqrt(vp1**-2 - p**2)
    qs1 = np.sqrt(vs1**-2 - p**2)
    qp2 = np.sqrt(vp2**-2 - p**2)
    qs2 = np.sqrt(vs2**-2 - p**2)
    
    dmu = mu1 - mu2
    drho = rho1 - rho2

    d1 = 2*p**2*dmu*(qp1 - qp2) + (rho1*qp2 + rho2*qp1)
    d2 = 2*p**2*dmu*(qs1 - qs2) + (rho1*qs2 + rho2*qs1)
    d3 = p*(2*dmu*(qp1*qs2 + p**2)-drho)
    d4 = p*(2*dmu*(qp2*qs1 + p**2)-drho)

    c1 = 2*p**2*dmu*(qp1 + qp2) - (rho1*qp2 - rho2*qp1)
    c2 = -2*p**2*dmu*(qs1 + qs2) - (rho1*qs2 - rho2*qs1)
    c3 = -p*(2*dmu*(qp1*qs2-p**2)+drho)
    c4 = -p*(2*dmu*(qp2*qs1-p**2)+drho)

    A2 = 4*p**2*rho2*vs2**4*qs2
    B = 1 - 2*p**2*vs2**2
    A1 = B**2
    C1 = 2*p*rho1*vs2**2
    C2 = 2*p*rho2*vs2**2

    # Solid/solid interface
    det = (d1*d2 + d4*d3)
    r_DPP = (c1*d2 - c3*d4)/det
    r_DPS = -(vp1/vs1)*(c3*d1 + c1*d3)/det
    r_DSP = (vs1/vp1)*(c4*d2 - c2*d4)/det
    r_DSS = -(c2*d1 + c4*d3)/det    
    
    t_DPP = (vp1/vp2)*(2*rho1*qp1*d2)/det
    t_DPS = (vp1/vs2)*(2*rho1*qp1*d4)/det # changed polarity of t_DPS formula in Amundsen's to fit Crewes 
    t_DSP = (vs1/vp2)*(2*rho1*qs1*d3)/det
    t_DSS = (vs1/vs2)*(2*rho1*qs1*d1)/det
    
    t_UPP = (vp2/vp1)*(2*rho2*qp2*d2)/det    
    
    # turn back on all warnings
    np.seterr(all="warn")    
    

    # make output dict
    out = {"r_DPP": r_DPP, "r_DPS": r_DPS, "r_DSP": r_DSP, "r_DSS": r_DSS, 
           "t_DPP": t_DPP, "t_DPS": t_DPS, "t_DSP": t_DSP, "t_DSS": t_DSS, 
           "t_UPP": t_UPP}
    
    # add NaN base layer (for consistent sizes)
    vv = np.nan * np.ones((1, nangles))
    for key in out:            
        out[key] = np.concatenate([out[key], vv])          
            
    return out
    
        


import numpy as np
from scipy.interpolate import interp1d


def insert_model_top(depth_min, depth, vp, vs, rhob, delta, epsilon, gamma):

    if len(np.array([depth_min])) == 0:
        return depth, vp, vs, rhob, delta, epsilon, gamma

    # double first layer (values defined as down-to)
    depth = np.insert(depth, 0, depth_min)
    if len(vp) > 0:
        vp = np.insert(vp, 0, vp[0])
    if len(vs) > 0:
        vs = np.insert(vs, 0, vs[0])
    if len(rhob) > 0:
        rhob = np.insert(rhob, 0, rhob[0])
    if len(delta) > 0:
        delta = np.insert(delta, 0, delta[0])
    if len(epsilon) > 0:
        epsilon = np.insert(epsilon, 0, epsilon[0])
    if len(gamma) > 0:
        gamma = np.insert(gamma, 0, gamma[0])

    return depth, vp, vs, rhob, delta, epsilon, gamma

def merge_layers(depth, vp, vs, rhob, delta=[], epsilon=[], gamma=[]):
    
    # loop through layers and merge if identical
    
    # input size
    nlayers = len(depth)
    

    # put input arrays togother in 2D array    
    M = np.vstack((vp, vs, rhob)).transpose()
    try:
        N = np.vstack((delta, epsilon, gamma)).transpose()
        M = np.hstack((M,N))
    except:
        pass
        
    # indices of layers that are redudant
    idelete = np.array([]).astype(int)        
            
    # loop
    for ii in np.arange(1,nlayers-1):
        
        top_layer = M[ii,:]
        bot_layer = M[ii+1,:]
        
        if np.array_equal(top_layer,bot_layer):
            idelete = np.append(idelete,ii)
    
    # remove redundant layers
    depth = np.delete(depth,idelete)
    vp = np.delete(vp,idelete)
    vs = np.delete(vs,idelete)
    rhob = np.delete(rhob,idelete)
    try: 
        delta = np.delete(delta,idelete)
        epsilon = np.delete(epsilon,idelete)
        gamma = np.delete(gamma,idelete)        
    except:
        pass
    
    return depth, vp, vs, rhob, delta, epsilon, gamma    

def remove_nan(depth, vp, vs, rhob, delta=[], epsilon=[], gamma=[]):

    # get input size
    nn = len(depth)

    # multiply together arrays to find all NaN-values
    vv = depth * vp * vs * rhob

    # include anisotropy
    if len(delta) == nn:
        vv = vv * delta
    if len(epsilon) == nn:
        vv = vv * epsilon
    if len(gamma) == nn:
        vv = vv * gamma

    # find nan's
    kk = np.where(np.isnan(vv) == False)

    # remove part with nan's
    depth = depth[kk]
    vp = vp[kk]
    vs = vs[kk]
    rhob = rhob[kk]

    # include anisotropy
    if len(delta) == nn:
        delta = delta[kk]
    if len(epsilon) == nn:
        epsilon = epsilon[kk]
    if len(gamma) == nn:
        gamma = gamma[kk]

    # return model without nan's
    return depth, vp, vs, rhob, delta, epsilon, gamma


def resample(x_old, y_old, x_new, kind="linear"):

    # reasmple single array
    ff = interp1d(x_old, y_old, kind=kind)
    return ff(x_new)


def depth_resample(
    dz, depth, vp, vs, rhob, delta=[], epsilon=[], gamma=[], kind="linear"
):

    # Resample model to uniform sampling (assume no nan-values)

    # output depth
    depth_new = np.arange(depth[0], depth[-1], dz)
    # from the arange docs: ...may result in the last element of out being greater than stop
    if depth_new[-1] > depth[-1]:
        depth_new = depth_new[:-1]

    # resample all model arrays
    vp = resample(depth, vp, depth_new, kind=kind)
    vs = resample(depth, vs, depth_new, kind=kind)
    rhob = resample(depth, rhob, depth_new, kind=kind)
    if len(delta) == len(depth):
        delta = resample(depth, delta, depth_new, kind=kind)
    if len(epsilon) == len(depth):
        epsilon = resample(depth, epsilon, depth_new, kind=kind)
    if len(gamma) == len(depth):
        gamma = resample(depth, gamma, depth_new, kind=kind)

    # return resampled model
    return depth_new, vp, vs, rhob, delta, epsilon, gamma


def depth_centre_to_layer(z_in, z_base=np.nan):

    # adjust depth from instantaneous (centered) to layer (down-to) values

    # make new depth column
    z_out = 0.5 * (z_in[1:] + z_in[:-1])

    # add base of last layer
    if np.isnan(z_base):
        z_base = z_out[-1] + np.diff(z_out)[-1]
    z_out = np.append(z_out, z_base)

    return z_out



import numpy as np
import matplotlib.pyplot as plt
from scipy.interpolate import interp1d
import base64
from io import BytesIO
import segyio
import lasio
from seismod1d.cmap import get_colormap


def resample_log(zz, val, zz_out):

    ff = interp1d(zz, val, kind="next", bounds_error=False, fill_value=np.nan)
    return ff(zz_out)


def to_las(
    output_file,
    depth_in,
    vp,
    vs,
    rhob,
    delta=[],
    epsilon=[],
    gamma=[],
    depth_step=0.1,
    null_value=-999.25,
):

    # create output depth curve
    depth = np.arange(depth_in[0], depth_in[-1] + depth_step, depth_step)
    depth = depth[depth <= depth_in[-1]]

    # initiate LAS-file
    las = lasio.LASFile()

    # header values
    las.well["ELEV"] = lasio.HeaderItem(
        "ELEV", unit="m", value="0", descr="SURFACE ELEVATION"
    )

    # values
    log_name = ["vp", "vs", "rhob", "delta", "epsilon", "gamma"]
    log_unit = ["m/s", "m/s", "g/cm3", "unitless", "unitless", "unitless"]
    log_descr = [
        "P-velocity",
        "S-velocity",
        "Bulk density",
        "Thomsen anisotropic delta",
        "Thomsen anisotropic epsilon",
        "Thomsen anisotropic gamma",
    ]
    nlog = len(log_name)

    # add depth curves to LAS
    las.add_curve("DEPTH", depth, unit="m", descr="Measured Depth")
    las.add_curve("TVD", depth, unit="m", descr="True Vertical Depth KB")
    las.add_curve("TVDMSL", depth, unit="m", descr="True Vertical Depth Mean Sea Level")

    # loop through data columns and add
    # las.add_curve('VP', resample_log(depth_in, vp, depth), unit='m/s', descr = 'P-velocity')
    for ii in np.arange(nlog):

        # get data
        val_in = eval(log_name[ii])

        # check if log is empty
        if len(val_in) > 0:

            # resample
            val = resample_log(depth_in, val_in, depth)

            # replace nan's
            val = np.nan_to_num(
                val, copy=True, nan=null_value, posinf=null_value, neginf=null_value
            )

            # add to las
            las.add_curve(
                log_name[ii].upper(), val, unit=log_unit[ii], descr=log_descr[ii]
            )

    # write file
    las.write(output_file, version=2)


def to_sgy(
    output_file, traces_x, traces_y, traces, il_const=1000, xl_const=2000,
):

    spec = segyio.spec()

    # issue warning if negative time/depth values
    if traces_y[0] < 0:
        print("WARNING: negative input sample values when writing segy file.")

    # get size of input
    ntr = len(traces_x)

    # specify structure
    spec = segyio.spec()
    spec.format = 1
    spec.sorting = 2
    spec.samples = traces_y
    spec.tracecount = ntr

    # add trace data
    with segyio.create(output_file, spec) as f:

        # loop over input traces
        for itr in np.arange(ntr):

            # trace header
            f.header[itr] = {
                segyio.su.offset: traces_x[itr],
                segyio.su.iline: il_const,
                segyio.su.xline: xl_const,
            }

            # trace data
            f.trace[itr] = traces[:, itr]

        f.bin.update(tsort=segyio.TraceSortingFormat.INLINE_SORTING)


def get_image_edges(xx, yy):

    # get image edges
    dy = yy[1] - yy[0]
    dx = xx[1] - xx[0]
    y_top = yy[0] - (dy / 2)
    y_base = yy[-1] + (dy / 2)
    x_left = xx[0] - (dx / 2)
    x_right = xx[-1] + (dx / 2)

    return [x_left, x_right, y_base, y_top]


def plot_image(
    traces_x,
    traces_y,
    traces,
    color_map_name="petrel_seismic_default",
    color_map_reverse=False,
    color_range_min=-0.3,
    color_range_max=0.3,
    int_method="bilinear",
    show_axis_ticks=False,
    upscaling_x=4,
    upscaling_y=8,
):

    # get colormap
    cmap = get_colormap(color_map_name, reverse=color_map_reverse)

    # parameters
    # int_method = "bilinear"
    sizes = np.shape(traces)

    # set figure size in pixels
    figure_width_pixels = upscaling_y * float(sizes[1])
    figure_height_pixels = upscaling_x * float(sizes[0])

    # create figure
    fig = plt.figure("SEISMIC TO PNG")
    fig.clf()

    # set figure size in inches
    figure_width_inches = figure_width_pixels / fig.dpi
    figure_height_inches = figure_height_pixels / fig.dpi
    fig.set_size_inches(figure_width_inches, figure_height_inches)

    # create axes
    ax = plt.Axes(fig, [0.0, 0.0, 1.0, 1.0])
    ax.set_axis_off()
    if show_axis_ticks:
        ax.set_axis_on()
        ax.tick_params(
            direction="in",
            length=6,
            width=1,
            colors="k",
            grid_color="k",
            grid_alpha=0.5,
        )
    ax.spines["top"].set_visible(False)
    fig.add_axes(ax)

    # get image edges
    extent = get_image_edges(traces_x, traces_y)

    # display image
    ax.imshow(
        traces,
        aspect="auto",
        cmap=cmap,
        vmin=color_range_min,
        vmax=color_range_max,
        interpolation=int_method,
        extent=extent,
    )

    # set axis limits
    plt.xlim(traces_x[0], traces_x[-1])
    plt.ylim(traces_y[-1], traces_y[0])


def get_image(
    traces_x,
    traces_y,
    traces,
    color_map_name="petrel_seismic_default",
    color_map_reverse=False,
    color_range_min=-0.3,
    color_range_max=0.3,
    int_method="bilinear",
    show_axis_ticks=False,
    upscaling_x=4,
    upscaling_y=8,
):

    # plot image
    plot_image(
        traces_x,
        traces_y,
        traces,
        color_map_name=color_map_name,
        color_map_reverse=color_map_reverse,
        color_range_min=color_range_min,
        color_range_max=color_range_max,
        int_method=int_method,
        show_axis_ticks=show_axis_ticks,
        upscaling_x=upscaling_x,
        upscaling_y=upscaling_y,
    )

    image_bytes = BytesIO()
    plt.savefig(image_bytes, format="png")
    image_bytes.seek(0)
    image_base64 = base64.b64encode(image_bytes.getvalue()).decode()

    return image_base64


def to_png(
    output_file,
    traces_x,
    traces_y,
    traces,
    color_map_name="petrel_seismic_default",
    color_map_reverse=False,
    color_range_min=-0.3,
    color_range_max=0.3,
    int_method="bilinear",
    show_axis_ticks=False,
    upscaling_x=4,
    upscaling_y=8,
):

    # plot image
    plot_image(
        traces_x,
        traces_y,
        traces,
        color_map_name=color_map_name,
        color_map_reverse=color_map_reverse,
        color_range_min=color_range_min,
        color_range_max=color_range_max,
        int_method=int_method,
        show_axis_ticks=show_axis_ticks,
        upscaling_x=upscaling_x,
        upscaling_y=upscaling_y,
    )

    # export to file
    plt.savefig(output_file, format="png")



import numpy as np
import matplotlib.pyplot as plt

from seismod1d.ray1d import trace, lib
from seismod1d.model import reflectivity


class Ray1D:
    def __init__(
        self,
        depth=[],
        vp=[],
        vs=[],
        rhob=[],
        delta=[],
        epsilon=[],
        gamma=[],
        theta_in=np.arange(90),
        method="isotropic",
    ):

        # layered model
        self.depth = depth
        self.vp = vp
        self.vs = vs
        self.rhob = rhob
        self.delta = delta
        self.epsilon = epsilon
        self.gamma = gamma

        # matrices with raytracer result
        self.offset = np.empty(0).astype("complex")
        self.time = np.empty(0).astype("complex")
        self.phase_angle = np.empty(0).astype("complex")
        self.group_angle = np.empty(0).astype("complex")
        self.phase_velocity = np.empty(0).astype("complex")
        self.group_velocity = np.empty(0).astype("complex")
        self.ref_coef = np.empty(0).astype("complex")

        # additional parameters
        self.theta_in = theta_in
        self.method = method

    def run_raytracer(self):

        # get input parameters
        depth = self.depth
        vp = self.vp
        vs = self.vs
        delta = self.delta
        epsilon = self.epsilon

        # avoid zero vs
        vs[vs == 0] = 1e-12

        # run raytracer
        # print('Running raytracer 1D, method: ' + self.method + ' ...')
        if self.method == "isotropic":
            (
                offset,
                time,
                phase_angle,
                group_angle,
                phase_velocity,
                group_velocity,
            ) = trace.iso(depth, vp, theta_in=self.theta_in)
        elif self.method == "vti":
            (
                offset,
                time,
                phase_angle,
                group_angle,
                phase_velocity,
                group_velocity,
            ) = trace.vti(depth, vp, vs, delta, epsilon, theta_in=self.theta_in)
        # print('...done')

        # assign output
        self.offset = offset
        self.time = time
        self.phase_angle = phase_angle
        self.group_angle = group_angle
        self.phase_velocity = phase_velocity
        self.group_velocity = group_velocity

    def resample_offset(self, offset_out):

        # resample raytracer to uniform offset sampling
        self.time = lib.resample_offset_single(self.offset, self.time, offset_out)
        self.phase_angle = lib.resample_offset_single(
            self.offset, self.phase_angle, offset_out
        )
        self.group_angle = lib.resample_offset_single(
            self.offset, self.group_angle, offset_out
        )
        self.phase_velocity = lib.resample_offset_single(
            self.offset, self.phase_velocity, offset_out
        )
        self.group_velocity = lib.resample_offset_single(
            self.offset, self.group_velocity, offset_out
        )
        self.ref_coef = lib.resample_offset_single(
            self.offset, self.ref_coef, offset_out
        )
        self.offset = np.tile(offset_out, (self.offset.shape[0], 1))

    def get_reflectivity(self, ref_method="isotropic", apply_tr_loss=False):

        # get reflectivity
        self.ref_coef = reflectivity.multi_layer(
            self.phase_angle,
            self.vp,
            self.vs,
            self.rhob,
            delta=self.delta,
            epsilon=self.epsilon,
            gamma=self.gamma,
            ref_method=ref_method,
            apply_tr_loss=apply_tr_loss
        )

    def display_raytracer(self):

        # define log types
        offset = self.offset
        time = self.time

        # get input size
        nlayers, nrays = offset.shape

        # make figure
        fig = plt.figure("Raytracer display")
        fig.clf()
        nsub = 1
        isub = 1

        # raytracer display
        plt.subplot(1, nsub, isub)
        isub = isub + 1

        # loop through layers and plot
        for ii in np.arange(nlayers):

            # get plot vectors
            xx = offset[ii, :]
            tt = time[ii, :]

            # plot
            plt.plot(xx, tt, ".-b")

        plt.grid()
        plt.xlabel("offset (m)")
        plt.ylabel("time (s)")
        plt.gca().invert_yaxis()

        # make additional figure
        fig = plt.figure("Raytracer display 2 ", figsize=(12, 12))
        fig.clf()
        nrow = 2
        ncol = 5
        isub = 1
        cmap = "gist_rainbow"
        vel_min = 0  # np.nanmin(self.phase_velocity)
        vel_max = self.phase_velocity[-1, 0]

        # time
        plt.subplot(nrow, ncol, isub)
        isub = isub + 1
        im = plt.imshow(
            self.time,
            vmin=0,
            vmax=np.nanmax(self.time[-1, :]),
            cmap=cmap,
            aspect="auto",
            interpolation="none",
        )
        plt.grid()
        plt.title("time (s)")
        fig.colorbar(im, ax=plt.gca())

        # offset
        plt.subplot(nrow, ncol, isub)
        isub = isub + 1
        im = plt.imshow(
            self.offset,
            vmin=0,
            vmax=np.nanmax(self.offset[-1, :]),
            cmap=cmap,
            aspect="auto",
            interpolation="none",
        )
        plt.grid()
        plt.title("offset (m)")
        fig.colorbar(im, ax=plt.gca())

        # phase angles
        plt.subplot(nrow, ncol, isub)
        isub = isub + 1
        im = plt.imshow(
            np.real(self.phase_angle),
            vmin=0,
            vmax=90,
            cmap=cmap,
            aspect="auto",
            interpolation="none",
        )
        plt.grid()
        plt.title("phase_angle (deg)")
        fig.colorbar(im, ax=plt.gca())

        # group angles
        plt.subplot(nrow, ncol, isub)
        isub = isub + 1
        im = plt.imshow(
            np.real(self.group_angle),
            vmin=0,
            vmax=90,
            cmap=cmap,
            aspect="auto",
            interpolation="none",
        )
        plt.grid()
        plt.title("group_angle (deg)")
        fig.colorbar(im, ax=plt.gca())

        # difference phase vs group angles
        plt.subplot(nrow, ncol, isub)
        isub = isub + 1
        im = plt.imshow(
            np.real(self.group_angle - self.phase_angle),
            cmap=cmap,
            aspect="auto",
            interpolation="none",
        )
        plt.grid()
        plt.title("group_angle - phase_angle")
        fig.colorbar(im, ax=plt.gca())

        # RC
        plt.subplot(nrow, ncol, isub)
        isub = isub + 1
        cr = 0.2
        if len(self.ref_coef) > 0:
            im = plt.imshow(
                np.real(self.ref_coef),
                vmin=-cr,
                vmax=cr,
                cmap=cmap,
                aspect="auto",
                interpolation="none",
            )
        plt.grid()
        plt.title("RC (real)")
        fig.colorbar(im, ax=plt.gca())

        # RC
        plt.subplot(nrow, ncol, isub)
        isub = isub + 1
        cr = 0.2
        if len(self.ref_coef) > 0:
            im = plt.imshow(
                np.imag(self.ref_coef),
                vmin=-cr,
                vmax=cr,
                cmap=cmap,
                aspect="auto",
                interpolation="none",
            )
        plt.grid()
        plt.title("RC (imag)")
        fig.colorbar(im, ax=plt.gca())

        # phase velocities
        plt.subplot(nrow, ncol, isub)
        isub = isub + 1
        im = plt.imshow(
            np.real(self.phase_velocity),
            vmin=vel_min,
            vmax=vel_max,
            cmap=cmap,
            aspect="auto",
            interpolation="none",
        )
        plt.grid()
        plt.title("phase_velocity (m/s)")
        fig.colorbar(im, ax=plt.gca())

        # group velocities
        plt.subplot(nrow, ncol, isub)
        isub = isub + 1
        im = plt.imshow(
            np.real(self.group_velocity),
            vmin=vel_min,
            vmax=vel_max,
            cmap=cmap,
            aspect="auto",
            interpolation="none",
        )
        plt.grid()
        plt.title("group_velocity (m/s)")
        fig.colorbar(im, ax=plt.gca())

        # difference phase vs group velocities
        plt.subplot(nrow, ncol, isub)
        isub = isub + 1
        im = plt.imshow(
            np.real(self.group_velocity - self.phase_velocity),
            cmap=cmap,
            aspect="auto",
            interpolation="none",
        )
        plt.grid()
        plt.title("group_velocity  - phase_velocity (m/s)")
        fig.colorbar(im, ax=plt.gca())



import numpy as np
from scipy.interpolate import interp1d


def get_time(zz, vp, start_time=0):

    # returns time by integrating velocity with depth
    # assumes constant velocity between depth points, i.e. vp[i] valid between zz[i-1] and zz[i]

    # calculate time
    dz = np.diff(zz)  # depth step (m)
    dt = 2 * dz / vp[1:]  # time step (s)
    tt = np.cumsum(dt)

    # insert zero time at top
    tt = np.insert(tt, 0, 0)

    # add start time
    tt = tt + start_time

    return tt


def get_time_linear(zz, vp):

    # returns time by integrating velocity with depth
    # assumes linear velocity between depth points

    # derive time
    dvp = np.diff(vp)
    dzz = np.diff(zz)
    ln_vp = np.log(vp)
    dln_vp = np.diff(ln_vp)
    aa = dvp / dzz

    # compute time step different if constant velocity across interval (avoid 0/0)
    dtt = 2 * dzz / vp[1:]
    ii = dvp != 0
    dtt[dvp != 0] = 2 * (aa[ii] ** -1) * dln_vp[ii]

    # sum time steps
    tt = np.cumsum(dtt)
    tt = np.insert(tt, 0, 0)

    # return time
    return tt


def get_time_from_td(z_td, t_td, z_in, kind="interp", int_method="linear"):

    # get time from td-realtion sampled at input depth
    if kind == "interp":

        ff = interp1d(
            z_td, t_td, kind=int_method, bounds_error=False, fill_value=np.nan
        )
        return ff(z_in)

    elif kind == "constant_vel":  # assumes constant velocity between input samples

        return 0


def get_depth_from_td(z_td, t_td, t_in, kind="interp", int_method="linear"):

    # get time from td-realtion sampled at input depth
    if kind == "interp":
        ff = interp1d(
            t_td, z_td, kind=int_method, bounds_error=False, fill_value="extrapolate"
        )
        return ff(t_in)


def apply_td_stretch(ref_time, model_depth, model_time, td_depth, td_time, direction=1):

    # zero-offset time
    ref_time0 = ref_time[:, 0]

    # get depth at reflection times from model time-depth

    ff = interp1d(
        model_time,
        model_depth,
        kind="linear",
        bounds_error=False,
        fill_value="extrapolate",
    )
    ref_depth0 = ff(ref_time0)

    # get time at reflection depths based on input td-relationship
    ff = interp1d(
        td_depth, td_time, kind="linear", bounds_error=False, fill_value="extrapolate"
    )
    ref_time_new0 = ff(ref_depth0)

    # get difference from current zero-offset time
    dt = ref_time_new0 - ref_time0

    # create 2D matrix of size ref_time based on dt
    nt = len(dt)
    ntr = ref_time.shape[1]  # number of output traces
    DT = np.tile(dt.reshape(nt, 1), (1, ntr))

    # return new time-table
    return ref_time + direction * DT



import numpy as np
import matplotlib.pyplot as plt
import scipy
from seismod1d import cmap

# import bravos.domainutils.reflectivity_vti as reflectivity_vti


def make_traces(RT, RC, ww_t, ww_a, tmin=[], tmax=[]):
    """
    Input reflection coefficients for multiple trace, convolve with wavelet and
    output synthetic seismogram

    Args:
        RT (2D array): Time of each reflection event. Columns are separate traces
        (angle/offset). Rows represents reflection surfaces.
        RC (2D array): Reflection coefficients. Same size as above array.
        ww_t (1D array): Wavelet time vector (assumes even sampling)
        ww_a (1D array): Wavelet amplitude vector
        tmin (value): optional output minimum time
        tmax (value): optional output maximum time
    Returns:
        TR (2D array): Synthetic seismogram.
    """

    # check if input is 1D array
    if RT.ndim == 1:
        RT = RT.reshape(RT.shape[0], -1)
        RC = RC.reshape(RC.shape[0], -1)

    # get shape of input
    nrc_y, nrc_x = np.shape(RC)

    # create time vector for convolution
    wav_duration = 2 * np.abs(ww_t[0])  # get wavelet duration
    dt = ww_t[1] - ww_t[0]  # use sampling from wavelet
    if np.array(tmin).size == 0:
        tmin = np.nanmin(RT) - 0.5 * wav_duration
    else:
        tmin = dt * np.floor(tmin / dt)
    if np.array(tmax).size == 0:
        tmax = np.nanmax(RT) + 0.5 * wav_duration
    else:
        tmax = dt * np.ceil(tmax / dt)
    tv = np.arange(tmin, tmax + dt, dt)  # make time vector for convolution
    ntv = len(tv)

    # define frequency vector for FFT
    nf = int(pow(2, np.ceil(np.log(ntv) / np.log(2))))
    df = 1 / (dt * nf)
    # frequency sampling
    fmin = -0.5 * df * nf
    fv = fmin + df * np.arange(0, nf)  # frequency vector

    # wavelet in f-domain
    ww_ap = np.zeros(nf)  # initilize wavelet amplitude padded with zeros
    ww_ap[: len(ww_a)] = ww_a  # insert wavelet amplitude
    ww_af = np.fft.fft(ww_ap)  # wavelet in frequency domain
    ww_af = np.fft.fftshift(ww_af)

    # corresponding padded time vector
    tv_p = tv[0] + ww_t[0] + dt * np.arange(0, nf)
    nt_p = len(tv_p)

    # initialize output
    TR = np.zeros([nt_p, nrc_x])
    for itr in np.arange(nrc_x):

        # make single trace (fourier domain)
        tr_f = conv_single_trace(
            RT[:, itr] - tv[0], RC[:, itr], fv, ww_af
        )  # subtract start time of time vector
        tr_f = np.fft.ifftshift(tr_f)
        tr_t = np.fft.ifft(tr_f)  # get trace in time domain
        TR[:, itr] = np.real(tr_t)  # take real value as output

    # crop traces to output time range (remove time range added for convolution)
    kk = (tv_p >= tv[0]) & (tv_p <= tv[-1])
    tv_out = tv_p[kk]
    TR = TR[kk, :]

    # return traces and time vector
    return TR, tv_out


def conv_single_trace(rc_t, rc_v, fv, ww_af):
    """
    Convolve a series of discrete reflection coeficients with a wavelet to
     obtain a seismic trace, by multiplying their frequency spectra.

    Args:
        rc_t (ndarray): The time of each reflection coefficient (s)
        rc_v (ndarray): The reflection coefficients (possibly complex values)
        fv (ndarray): Frequency vector
        ww_af (ndarray): The wavelet in frequency domain
    Returns:
        tr_f: (ndarray): Seismic trace in frequency domain

    """

    # initiate output
    tr_f = np.zeros(len(fv))

    # remove NaN values
    kk = np.isnan(rc_t * rc_v)

    # remove NaN values
    rc_t = np.delete(rc_t, kk)
    rc_v = np.delete(rc_v, kk)

    # return if all NaN
    if len(rc_t) == 0:
        return tr_f

    # reshape vectors
    fv = fv.reshape(fv.shape[0], -1)  # column vector
    rc_t = rc_t.reshape(1, rc_t.shape[0])  # row vector
    rc_v = rc_v.reshape(rc_v.shape[0], -1)  # column vector

    # convolve (multiplication sum in fourier domain)
    tr_f = ww_af * np.matmul(np.exp(-1j * 2 * np.pi * np.matmul(fv, rc_t)), rc_v)[:, 0]

    return tr_f


def apply_phase_rot(TR):

    phi = -np.pi / 2  # rotation angle
    ncol = TR.shape[1]  # number of columns

    for ii in np.arange(ncol):

        TR[:, ii] = _rotate_phase(TR[:, ii], phi)

    return TR


def display_traces(xx, tt, traces, xlim=[], ylim=[], xlabel=""):

    # if input is 1D, double traces before display
    if len(xx) == 1:
        traces = np.hstack((traces, traces))
        xx = np.array([xx[0] - 1, xx[0] + 1])

    # get petrel colormap
    color_map = cmap.get_colormap("petrel_seismic_default", reverse=True)

    # make additional figure
    fig = plt.figure("Synthetics ", figsize=(12, 12))
    fig.clf()
    nrow = 1
    ncol = 1
    isub = 1

    # image properties
    # cmap = 'seismic'
    cr = 0.3
    int_method = "bilinear"
    # int_method = 'bicubic'
    # int_method = 'none'

    # calcualte image edges
    dt = np.diff(tt)[0]
    dx = np.diff(xx)[0]
    t_top = tt[0] - (dt / 2)
    t_base = tt[-1] + (dt / 2)
    x_left = xx[0] - (dx / 2)
    x_right = xx[-1] + (dx / 2)
    extent = [x_left, x_right, t_base, t_top]

    # display image
    plt.subplot(nrow, ncol, isub)
    im = plt.imshow(
        traces,
        aspect="auto",
        cmap=color_map,
        vmin=-cr,
        vmax=cr,
        interpolation=int_method,
        extent=extent,
    )

    if len(np.array(ylim)) == 0:
        ylim = (tt[-1], tt[0])

    if len(np.array(xlim)) == 0:
        xlim = (xx[0], xx[-1])

    # plot settings
    plt.xlim(xlim)
    plt.ylim(ylim)
    plt.grid()
    plt.ylabel("Time (s)")
    plt.xlabel(xlabel)

    # show colorbar
    fig.colorbar(im, ax=plt.gca())

    # # synthetics
    # plt.subplot(nrow, ncol, isub); isub = isub + 1
    # im = plt.imshow(self.time, vmin = 0, vmax = np.nanmax(self.time[-1,:]), cmap = cmap, aspect='auto', interpolation='none')

    # plt.title('time (s)')
    # fig.colorbar(im, ax = plt.gca())


def _apply_along_axis(func_1d, arr, *args, **kwargs):
    """
    Apply 1D function across 2D slice as efficiently as possible.

    Although `np.apply_along_axis` seems to do well enough, map usually
    seems to end up beig a bit faster.

    Args:
        func_1d (function): the 1D function to apply, e.g. np.convolve. Should
            take 2 or more arguments: the

    Example
    >>> apply_along_axes(np.convolve, reflectivity_2d, wavelet, mode='same')
    """
    mapobj = map(lambda tr: func_1d(tr, *args, **kwargs), arr)
    return np.array(list(mapobj))


def _rotate_phase(w, phi, degrees=False):
    r"""
    Performs a phase rotation of wavelet or wavelet bank using:

    .. math::

        A = w(t)\cos\phi - h(t)\sin\phi

    where `w(t)` is the wavelet, `h(t)` is its Hilbert transform, and \phi is
    the phase rotation angle (default is radians).

    The analytic signal can be written in the form :math:`S(t) = A(t)e^{j\theta (t)}`
    where :math:`A(t) = \left| h(w(t)) \right|` and :math:`\theta(t) = \tan^{-1}[h(w(t))]`. 
    `A(t)` is called the "reflection strength" and :math:`\phi(t)` is called the "instantaneous
    phase".

    A constant phase rotation :math:`\phi` would produce the analytic signal
    :math:`S(t)=A(t)e^{j(\theta(t) + \phi)}`. To get the non-analytic signal,
    we take 

    .. math::

        real(S(t)) &= A(t)\cos(\theta(t) + \phi) \\
        &= A(t)\cos\theta(t)\cos(\phi)-\sin\theta(t)\sin(\phi))\\
        &= w(t)\cos\phi-h(t)\sin\phi
        

    Args:
        w (ndarray): The wavelet vector, can be a 2D wavelet bank.
        phi (float): The phase rotation angle (in radians) to apply.
        degrees (bool): If phi is in degrees not radians.

    Returns:
        The phase rotated signal (or bank of signals).
    """
    # Make sure the data is at least 2D to apply_along
    data = np.atleast_2d(w)

    # Get Hilbert transform. This will be 2D.
    a = _apply_along_axis(scipy.signal.hilbert, data, axis=0)

    # Transform angles into what we need.
    phi = np.asanyarray(phi).reshape(-1, 1, 1)
    if degrees:
        phi = np.radians(phi)

    rotated = np.real(a) * np.cos(phi) - np.imag(a) * np.sin(phi)
    return np.squeeze(rotated)



import numpy as np
import pandas as pd
import time
from seismod1d.wavelet import get_wavelet
from seismod1d.model import lib, reflectivity, synthetics
from seismod1d import td
from seismod1d.rockphysics import backus
from seismod1d.ray1d.ray1d import Ray1D
from seismod1d.ray1d import nmo
from seismod1d.ray1d.lib import resample_to_uniform_offset
from seismod1d.model import reflectivity_vti, reflectivity_iso  # , reflectivity_vti_test


NSPACE = 4


class SynthModel:
    def __init__(
        self,
        depth,
        vp,
        vs,
        rhob,
        delta=[],
        epsilon=[],
        gamma=[],
        depth_top=[],
        time_top=0,
    ):

        # if top depth is defined, create top layer
        if depth_top != []:
            depth, vp, vs, rhob, delta, epsilon, gamma = lib.insert_model_top(
                depth_top, depth, vp, vs, rhob, delta, epsilon, gamma
            )

        # layered model
        self.depth = np.array(depth, dtype=np.float64)
        self.vp = np.array(vp, dtype=np.float64)
        self.vs = np.array(vs, dtype=np.float64)
        self.rhob = np.array(rhob, dtype=np.float64)
        self.delta = np.array(delta, dtype=np.float64)
        self.epsilon = np.array(epsilon, dtype=np.float64)
        self.gamma = np.array(gamma, dtype=np.float64)

        # top of model
        self.time_top = time_top
        self.time = td.get_time(self.depth, self.vp, start_time=self.time_top)

        # general model settings
        self.model_domain = "angle"  # model in angle or offset
        self.model_type = "isotropic"  # isotropic, vti
        self.model_angle = np.arange(0, 90)
        dz = 10
        self.model_offset = np.arange(0, 6000 + dz, dz)
        # self.model_offset_type = 'moveout' # moveout, nmo-corr

        # reflectivity
        self.ref_coef = np.empty(0).astype("complex")  # reflectivity coefficient
        self.ref_time = np.empty(0).astype("float")  # reflectivity time (s)

        # trace data
        self.traces = []  # 2D (N x M) array with synthetic trace data
        self.traces_time = []  # 1D array (N) with time values (uniform sampling)
        self.traces_x = []  # 1D array (M) with angle/offset values
        self.traces_depth = []  # 1D array (N) with depth values (non-uniform sampling)

        # raytracer
        self.raytracer = Ray1D()

    def precondition(self, 
                     depth_step_resample=0.1, 
                     apply_value_centering=False,
                     depth_crop_after_centering=False,
                     apply_depth_resample=True,
                     apply_backus=True,
                     merge_layers=True,
                     backus_step=2, 
                     backus_qc_plot=False,                     
                     ):

        # measure time
        start_time = time.time()        

        # remove leading/trailing nan's
        print("Removing leading/trailing nan values ...")
        (
            self.depth,
            self.vp,
            self.vs,
            self.rhob,
            self.delta,
            self.epsilon,
            self.gamma,
        ) = lib.remove_nan(
            self.depth,
            self.vp,
            self.vs,
            self.rhob,
            self.delta,
            self.epsilon,
            self.gamma,
        )        

        # adjust depth from centered to down-to definition
        if apply_value_centering:
            print("Adjusting depths from centered definition to layered (down-to)...")
            if depth_crop_after_centering:
                self.depth = lib.depth_centre_to_layer(self.depth, z_base=self.depth[-1])
            else:
                self.depth = lib.depth_centre_to_layer(self.depth)

        # resample to uniform depth sampling
        if apply_depth_resample:
            print("Resampling to uniform depth sampling (dz = " + str(depth_step_resample) + ")...")
            (
                self.depth,
                self.vp,
                self.vs,
                self.rhob,
                self.delta,
                self.epsilon,
                self.gamma,
            ) = lib.depth_resample(
                depth_step_resample,
                self.depth,
                self.vp,
                self.vs,
                self.rhob,
                self.delta,
                self.epsilon,
                self.gamma,
            )                    

        # backus blocking
        if apply_backus: 
            print("Backus blocking...")
            start_time = time.time()
            (
                self.depth,
                self.vp,
                self.vs,
                self.rhob,
                self.delta,
                self.epsilon,
                self.gamma,
            ) = backus.constant_length(
                self.depth,
                self.vp,
                self.vs,
                self.rhob,
                self.delta,
                self.epsilon,
                self.gamma,
                block_length=backus_step,
                qc_plot=backus_qc_plot,
            )
            print(
                "...done. Time elapsed backus: "
                + "{:.3f}".format(time.time() - start_time)
                + " s"
            )    
            
        # merge layers with identical elastic parameters
        if merge_layers:
            print("Merging identical layers ...")            
            (
                self.depth,
                self.vp,
                self.vs,
                self.rhob,
                self.delta,
                self.epsilon,
                self.gamma,
            ) = lib.merge_layers(                
                self.depth,
                self.vp,
                self.vs,
                self.rhob,
                self.delta,
                self.epsilon,
                self.gamma,
            )                               

        # recalculate model time
        self.time = td.get_time(self.depth, self.vp, start_time=self.time_top)

    def info(self):

        # add to pandas dataframe and print header
        df = pd.DataFrame(
            {"DEPTH": self.depth, "VP": self.vp, "VS": self.vs, "RHOB": self.rhob}
        )

        # add VTI anisotropy columns
        (nval, ncol) = df.shape
        if len(self.delta) == 0:
            self.delta = np.zeros(nval)
        if len(self.epsilon) == 0:
            self.epsilon = np.zeros(nval)
        if len(self.gamma) == 0:
            self.gamma = np.zeros(nval)

        df["DEL"] = self.delta
        df["EPS"] = self.epsilon
        df["GAM"] = self.gamma

        print(df.head())
        print(df.tail())

    def get_reflectivity(self, apply_tr_loss = False, raytracer_angles=np.arange(0, 90)):

        # determine reflection and raytracer calculation method
        if self.model_type == "isotropic":
            ray_method = "isotropic"
            ref_method = "zoeppritz_rpp"
        elif self.model_type == "vti":
            ray_method = "vti"
            ref_method = "vti_exact_rpp"

        # get reflection times
        if self.model_domain == "angle":  # angle domain output (AVA-modelling)

            # get reflectivity
            print("Get reflectivity of all layers: " + ref_method + "...")
            start_time = time.time()
            self.ref_coef = reflectivity.multi_layer(
                self.model_angle,
                self.vp,
                self.vs,
                self.rhob,
                delta=self.delta,
                epsilon=self.epsilon,
                gamma=self.gamma,
                ref_method=ref_method,
                apply_tr_loss=apply_tr_loss,
            )
            print(
                "...done. Time elapsed, reflectivity: "
                + "{:.3f}".format(time.time() - start_time)
                + " s"
            )

            # get reflectivity time (zero-offset time from model)
            tv = self.time

            # make 2D array of reflection time
            self.ref_time = np.tile(
                tv.reshape(tv.shape[0], 1), (1, self.ref_coef.shape[1])
            )

            # set model x
            self.traces_x = self.model_angle

        elif self.model_domain == "offset":  # angle domain output (AVA-modelling)

            # run raytracer
            print("Running raytracer: " + ray_method + "...")
            start_time = time.time()
            self.raytracer = Ray1D(
                depth=self.depth,
                vp=self.vp,
                vs=self.vs,
                rhob=self.rhob,
                delta=self.delta,
                epsilon=self.epsilon,
                gamma=self.gamma,
                theta_in=raytracer_angles,
                method=ray_method,
            )
            self.raytracer.run_raytracer()
            print(
                "...done. Time elapsed, raytracer: "
                + "{:.3f}".format(time.time() - start_time)
                + " s"
            )

            # get reflectivity (added to raytracer)
            print("Get reflectivity of all layers: " + ref_method + "...")
            start_time = time.time()
            self.raytracer.get_reflectivity(ref_method=ref_method, apply_tr_loss=apply_tr_loss)
            print(
                "...done. Time elapsed, reflectivity: "
                + "{:.3f}".format(time.time() - start_time)
                + " s"
            )

            # resample reflectivity to uniform offset
            self.ref_coef = resample_to_uniform_offset(
                self.raytracer.offset, self.raytracer.ref_coef, self.model_offset
            )
            self.ref_time = resample_to_uniform_offset(
                self.raytracer.offset, self.raytracer.time, self.model_offset
            )

            # assign to model
            self.traces_x = self.model_offset

    def get_traces(
        self,
        wav_time=[],
        wav_amp=[],
        wav_dt=0.004,        
        time_min=[],
        time_max=[],        
        apply_nmo_stretch=False,
        apply_td_stretch=False,
        td_depth=[],
        td_time=[],
    ):

        # measure time
        start_time = time.time()

        # get default wavelet if empty
        if len(wav_time) == 0:
            print("Get default wavelet...")
            wav_time, wav_amp, wav_name = get_wavelet(wav_dt=wav_dt)

        # apply nmo stretch only in offset domain
        if self.model_domain == "angle":
            apply_nmo_stretch = False

        # if NMO is to be applied later, recalculate moveout curves upfront using NMO method (not raytracer)
        # this ensures event flatness as raytracer and NMO correction movout might differ slightly
        if apply_nmo_stretch:
            print("Recalculating moveout curves...")

            # establish temporary time-depth trend to use for NMO moveout estimation
            dt = wav_time[1] - wav_time[0]
            tv = np.arange(self.time[0], self.time[-1] + dt, dt)
            zv = td.get_depth_from_td(
                self.depth, self.time, tv, kind="interp", int_method="linear"
            )

            # get reflection time from NMO method
            self.ref_time = nmo.get(
                tv,
                zv,
                self.traces_x,
                self.depth,
                self.vs,
                self.delta,
                self.epsilon,
                nmo_type=self.model_type,
                tv_output=self.time,
                direction=1,
                nterm=4,
            )

        # apply td stretch prior to convolution with wavelet
        if apply_td_stretch:
            print("Applying TD stretch......")
            self.ref_time = td.apply_td_stretch(
                self.ref_time, self.depth, self.time, td_depth, td_time
            )

        # convolve with wavelet to make traces
        print("Making traces (convolution with wavelet)...")
        self.traces, self.traces_time = synthetics.make_traces(
            self.ref_time,
            self.ref_coef,
            wav_time,
            wav_amp,
            tmin=time_min,
            tmax=time_max,
        )
        print(
            "...done. Time elapsed, make traces: "
            + "{:.3f}".format(time.time() - start_time)
            + " s"
        )

        # apply NMO stretch
        if apply_nmo_stretch:
            print("Applying NMO stretch.....")

            # get depth at trace time
            zv = td.get_depth_from_td(
                self.depth,
                self.time,
                self.traces_time,
                kind="interp",
                int_method="linear",
            )

            # get reflection time from NMO method
            TX = nmo.get(
                self.traces_time,
                zv,
                self.traces_x,
                self.depth,
                self.vs,
                self.delta,
                self.epsilon,
                nmo_type=self.model_type,
                direction=1,
                nterm=4,
            )

            # apply td stretch
            if apply_td_stretch:

                # update nmo traveltime matrix according to td-stretch
                TX = td.apply_td_stretch(
                    TX, self.depth, self.time, td_depth, td_time
                )  # , direction = 1)
                TX = nmo.resample_tx(TX, self.traces_time)

            # apply NMO correction
            self.traces = nmo.apply_tx(self.traces, self.traces_time, TX)
            print("...done")

    def display_traces(self, xlim=[], ylim=[]):

        # make xlabel
        xlabel = "Angle (deg)"
        if self.model_domain == "offset":
            xlabel = "Offset (m)"

        # show traces
        synthetics.display_traces(
            self.traces_x,
            self.traces_time,
            self.traces,
            xlim=xlim,
            ylim=ylim,
            xlabel=xlabel,
        )



import numpy as np
import pandas as pd
from matplotlib.colors import ListedColormap
import pathlib


def get_colormap(cmap_name, reverse=False):

    # make absolute path
    bundle_dir = pathlib.Path(__file__).parent
    cmap_file = (bundle_dir / "colormaps" / (cmap_name + ".dat")).absolute()

    # read ASCII file
    pp = pd.read_csv(cmap_file)
    mm = pp.values
    nc = len(mm)

    # create colormap
    kk = np.ones(nc).reshape(nc, 1)
    newcolors = np.concatenate((mm, kk), axis=1)
    cmap0 = ListedColormap(newcolors)

    # reverse colormap
    if reverse:
        cmap0.colors = np.flip(cmap0.colors, axis=0)

    # retrn colormap
    return cmap0
